/* ***********************************************************************/
/*                                                                       */
/*                     WiNRADiO G69DDC EXCELSIOR ULTRA                   */
/*                  Header file for G69DDC device driver                 */
/*                      (c) Copyright 2022 RADIXON.                      */
/*                                                                       */
/* ***********************************************************************/
#ifndef __G69DDC_H__
#define __G69DDC_H__

#ifdef HAVE_CONFIG_H
#include <linux/config.h>
#endif

#define G69DDC_USB
//#define G69DDC_PCI
//#define G69DDC_DEBUG

#if defined(CONFIG_COMPAT)
#define G69DDC_COMPAT
#endif

#include <linux/version.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/time.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/kthread.h>
#include <linux/timer.h>
#include <linux/mutex.h>
#include <linux/uaccess.h>
#include <linux/scatterlist.h>
#include <linux/vmalloc.h>

#ifdef G69DDC_COMPAT
#include <linux/compat.h>
#endif //G69DDC_COMPAT


#ifdef G69DDC_PCI
#include <linux/pci.h>
#include <linux/interrupt.h>
#endif //G69DDC_PCI

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a<<16)|(b<<8)|c)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(4,0,0)
#pragma GCC error  "Your kernel version is not supported! Kernel v4.0 and above is required."
#endif

#if LINUX_VERSION_CODE>=KERNEL_VERSION(4,11,0)
#include <linux/sched/signal.h>
#endif

#include "g69ddc_ioctl.h"

#ifdef G69DDC_COMPAT
#include "g69ddc_ioctl_compat.h"
#endif //G69DDC_COMPAT

#define G69DDC_AUTHOR                   "RADIXON"
#define G69DDC_MODULE_NAME              "g69ddc"
#define G69DDC_MAX_DEVICE_COUNT         32
#define G69DDC_DESCRIPTION              "WiNRADiO G69DDC device driver"


struct g69ddc;

typedef void (*g69ddc_fnc_delete)(struct g69ddc *dev);
typedef void (*g69ddc_fnc_uninitialize)(struct g69ddc *dev);
typedef long (*g69ddc_fnc_ioctl)(struct g69ddc *dev,struct file *file, unsigned int cmd,unsigned long arg);
typedef int (*g69ddc_fnc_open)(struct g69ddc *dev);
typedef void (*g69ddc_fnc_close)(struct g69ddc *dev);
typedef int (*g69ddc_fnc_set_power)(struct g69ddc *dev,uint8_t power);
typedef int (*g69ddc_fnc_check_temperature)(struct g69ddc *dev,bool *critical_temperature);


struct g69ddc
{
    struct device                   dev;
    struct cdev                     cdev;
    uint32_t                        index;
    struct device                   *p_dev;
    struct mutex                    mutex;
    uint8_t                         bus_type;
    uint16_t                        vid_pid[4];
    uint16_t                        fw[4];
    bool                            open;
    union g69ddc_state              state;
    wait_queue_head_t               state_wait_queue;
    struct workqueue_struct         *workqueue;
    struct delayed_work             wd_delayed_work;
    uint32_t                        wd_counter;

    g69ddc_fnc_uninitialize         fnc_uninitialize;
    g69ddc_fnc_delete               fnc_delete;
    g69ddc_fnc_ioctl                fnc_ioctl;

#ifdef G69DDC_COMPAT
    g69ddc_fnc_ioctl                fnc_ioctl_compat;
#endif

    g69ddc_fnc_open                 fnc_open_locked;
    g69ddc_fnc_close                fnc_close_locked;
    g69ddc_fnc_set_power            fnc_set_power_locked;
    g69ddc_fnc_check_temperature    fnc_check_temperature_locked;
};

#define print_error(__fmt__,...) printk(KERN_ERR G69DDC_MODULE_NAME ": " __fmt__ "\n",##__VA_ARGS__)
#define print_info(__fmt__,...) printk(KERN_INFO G69DDC_MODULE_NAME ": " __fmt__ "\n",##__VA_ARGS__)
#define print_warn(__fmt__,...) printk(KERN_WARNING G69DDC_MODULE_NAME ": " __fmt__ "\n",##__VA_ARGS__)

#define dev_print_error(__dev__,__fmt__,...) printk(KERN_ERR G69DDC_MODULE_NAME ":%s: " __fmt__ "\n",dev_name(__dev__),##__VA_ARGS__)
#define dev_print_info(__dev__,__fmt__,...) printk(KERN_INFO G69DDC_MODULE_NAME ":%s: " __fmt__ "\n",dev_name(__dev__),##__VA_ARGS__)
#define dev_print_warn(__dev__,__fmt__,...) printk(KERN_WARNING G69DDC_MODULE_NAME ":%s: " __fmt__ "\n",dev_name(__dev__),##__VA_ARGS__)

#ifdef G69DDC_DEBUG

#define print_dbg(__fmt__,...) printk(KERN_DEBUG G69DDC_MODULE_NAME ": " __fmt__ "\n",##__VA_ARGS__)
#define dev_print_dbg(__dev__,__fmt__,...) printk(KERN_DEBUG G69DDC_MODULE_NAME ":%s: " __fmt__ "\n",dev_name(__dev__),##__VA_ARGS__)
#else

#define print_dbg(__fmt__,...)
#define dev_print_dbg(__dev__,__fmt__,...)

#endif //G69DDC_DEBUG

#ifdef G69DDC_USB

#include <linux/usb.h>

#define G69DDC_USB_TIMEOUT              5000
#define G69DDC_USB_EP_MIN_TRANSFER_SIZE 8192
#define G69DDC_USB_EP_MAX_TRASNFER_SIZE (8*1024*1024)

struct g69ddc_usb;

struct g69ddc_usb_ep
{
    struct g69ddc_usb   *dev;
    uint8_t             address;
    uint16_t            max_packet_size;
    uint8_t             interval;
    uint8_t             type;
    unsigned int        pipe;
    struct usb_anchor   anchor;
    spinlock_t          spinlock;
    struct list_head    buffer_list;
    wait_queue_head_t   wait_queue;
};

struct g69ddc_usb_ep_buffer
{
    struct g69ddc_usb_ep    *ep;
    struct list_head        list_entry;
    uintptr_t               user_buffer;
    uint32_t                size;
    struct urb              *urb;
    int                     page_count;
    struct page             **pages;
    void                    *vmem;
    bool                    idle;
    int                     status;
    uint32_t                bytes_transferred;
    struct scatterlist      *sg;
};

struct g69ddc_usb
{
    struct g69ddc               base_dev;
    struct usb_device           *usb_dev;
    struct usb_interface        *usb_interface;
    uint32_t                    led_mode;
    struct rw_semaphore         rw_sem;

    unsigned int                snd_ctrl_pipe;
    unsigned int                rcv_ctrl_pipe;

    uint32_t                    ep_count;
    struct g69ddc_usb_ep        *ep;
};

#endif //G69DDC_USB
#endif //__G69DDC_H__
